class InvalidOpenAIModel(Exception):
    pass


class InvalidWriterModel(Exception):
    pass


class MissingPromptTemplate(Exception):
    pass


class UnsupportedVectorStore(Exception):
    pass


class MissingUseIndex(Exception):
    pass


class UnsupportedLLM(Exception):
    pass


class InvalidPromptTemplate(Exception):
    pass
