__title__ = "MindsDB PyPI handler"
__package_name__ = "mindsdb_pypi_handler"
__version__ = "0.1.0"
__description__ = "MindsDB handler for PyPI"
__author__ = "Sadra Yahyapour"
__github__ = "https://github.com/mindsdb/mindsdb"
__pypi__ = "https://pypi.org/project/mindsdb/"
__license__ = "MIT"
__copyright__ = "Copyright 2023 - mindsdb"
